/**************************************************************************
 * FreeDOS 32 FAT Driver                                                  *
 * by Salvo Isaja                                                         *
 *                                                                        *
 * Copyright (C) 2001-2002, Salvatore Isaja                               *
 *                                                                        *
 * This is "fat.h" - Driver's constants, structures and prototypes        *
 *                                                                        *
 *                                                                        *
 * This file is part of the FreeDOS 32 FAT Driver.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is free software; you can redistribute it    *
 * and/or modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2 of the  *
 * License, or (at your option) any later version.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is distributed in the hope that it will be   *
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 * GNU General Public License for more details.                           *
 *                                                                        *
 * You should have received a copy of the GNU General Public License      *
 * along with the FreeDOS 32 FAT Driver; see the file COPYING;            *
 * if not, write to the Free Software Foundation, Inc.,                   *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                *
 **************************************************************************/

#ifndef __FAT1_H
#define __FAT1_H

#define FATBUFFERS /* Uncomment this to use the bufferized I/O */
#define FATLFN     /* Uncomment this to use Long File Names */

typedef unsigned long  DWORD;
typedef unsigned short WORD;
typedef unsigned char  BYTE;
typedef unsigned long  TIME;
typedef unsigned long  SYS_FLAGS;

typedef signed   char      SBYTE;
typedef signed   short     SWORD;
typedef signed   long      SDWORD;
typedef signed   long      SQWORD; //NOrmalerweise 8 Byte Integer!!!
typedef unsigned long      QWORD;  //NOrmalerweise 8 Byte Integer!!!


#define FD32_MAX_LFN_PATH_LENGTH 260
#define FD32_MAX_LFN_LENGTH      255
#define FD32_MAX_SFN_PATH_LENGTH 64
#define FD32_MAX_SFN_LENGTH      14

//#define NULL 0


#include "filesys.h"
#include "datetime.h"

typedef struct {
  int            active, type;
  unsigned long  start, size;
} Partition;


/* FAT Types */
#define FATTYPE_FAT12 0
#define FATTYPE_FAT16 1
#define FATTYPE_FAT32 2

//#define FAT_OPEN_USE_LFN          0x010000 /* Use long file names           */
#define FAT_OPEN_DIRECTORY        0x8000000 //(1 << 31) /* Open a directory as a file    */
/* Internal file modes to append to the opening mode for full file status */
#define FAT_FMODE_OPENED          0x0200000 /* The file is opened           */
#define FAT_FMODE_ROOT            0x0400000 /* File is a FAT12/FAT16 root   */
#define FAT_FMODE_NEGPOS          0x0800000 /* File target position is < 0  */
#define FAT_FMODE_FIRST_OPEN      0x1000000 /* Handle of first time opening */
#define FAT_FMODE_DO_WRITE        0x2000000 /* Doing a write operation      */

/* EOC (End Of Clusterchain) macros.                                     */
/* These expressions are true (nonzero) if the value of a FAT entry is   */
/* an EOC for the FAT type. An EOC indicates the last cluster of a file. */
#define  FAT12_EOC(EntryValue)  (EntryValue >= 0x0FF8)
#define  FAT16_EOC(EntryValue)  (EntryValue >= 0xFFF8)
#define  FAT32_EOC(EntryValue)  (EntryValue >= 0x0FFFFFF8)

/* Bad cluster marks.                                                    */
/* Set a FAT entry to the FATxx_BAD value to mark the cluster as bad.    */
/*                                                                       */
/* The FAT file system specification says that to avoid confusion, no    */
/* FAT32 volume should ever be configured such that 0x0FFFFFF7 is an     */
/* allocatable cluster number. In fact an entry that would point to the  */
/* cluster 0x0FFFFFF7 would be recognised as Bad instead. Since values   */
/* greater or equal than 0x0FFFFFF8 are interpreted as EOC, I think we   */
/* can assume that the max cluster for a FAT32 volume is 0x0FFFFFF6.     */
/* That problem doesn't exist on FAT12 and FAT16 volumes, in fact:       */
/* 0x0FF7 = 4087 is greater than 4086 (max cluster for a FAT12 volume)   */
/* 0xFFF7 = 65527 is greater than 65526 (max cluster for a FAT16 volume) */
#define FAT12_BAD 0x0FF7
#define FAT16_BAD 0xFFF7
#define FAT32_BAD 0x0FFFFFF7

/* Special codes for the first byte of a directory entry */
#define FAT_DIR_FREE     0xE5 /* The directory entry is free             */
#define FAT_DIR_ENDOFDIR 0x00 /* This and the following entries are free */
#define FAT_DIR_E5       0x05 /* The character is actually a chr 0xE5    */

/* FAT Return Codes                                          */
/* FIX ME: These should be removed. Use FD32 errors instead. */
#define FAT_RET_SECTOR_ERROR       -0xF0
#define FAT_RET_RMDIR_ROOT         -0xF1
#define FAT_RET_RMDIR_NOT_EMPTY    -0xF2
#define FAT_RET_NOT_A_DIRECTORY    -0xF3
#define FAT_RET_DIRECTORY_FULL     -0xF4
#define FAT_RET_INVALID_FILENAME   -0xF5
#define FAT_RET_INVALID_CLUSTER    -0xF6
#define FAT_RET_EOF                -0xF7
#define FAT_RET_SHARING_CRITICAL   -0xF8
#define FAT_RET_INVALID_SHARING    -0xF9
#define FAT_RET_CRITICAL_ERROR     -0xFF


/* FAT32 Boot Sector and BIOS Parameter Block                              */
/* This structure is in all driver functions, even if the volume is FAT12  */
/* or FAT16. In fact some FAT32 fields are checked anyway. So we load the  */
/* boot sector from the disk, we detect the FAT type and if it's a FAT32,  */
/* we fill the following struct as is, while if the volume is FAT12/FAT16  */
/* we copy in the right position the appropriate (common) fields.          */
typedef struct
{
  BYTE  BS_jmpBoot[3];
  BYTE  BS_OEMName[8];
  WORD  BPB_BytsPerSec;
  BYTE  BPB_SecPerClus;
  WORD  BPB_ResvdSecCnt;
  BYTE  BPB_NumFATs;
  WORD  BPB_RootEntCnt;
  WORD  BPB_TotSec16;
  BYTE  BPB_Media;
  WORD  BPB_FATSz16;
  WORD  BPB_SecPerTrk;
  WORD  BPB_NumHeads;
  DWORD BPB_HiddSec;
  DWORD BPB_TotSec32;
  /* Here start the FAT32 specific fields (offset 36) */
  DWORD BPB_FATSz32;
  WORD  BPB_ExtFlags;
  WORD  BPB_FSVer;
  DWORD BPB_RootClus;
  WORD  BPB_FSInfo;
  WORD  BPB_BkBootSec;
  BYTE  BPB_Reserved[12];
  /* The following fields are present also in a FAT12/FAT16 BPB,     */
  /* but at offset 36. In a FAT32 BPB they are at offset 64 instead. */
  BYTE  BS_DrvNum;
  BYTE  BS_Reserved1;
  BYTE  BS_BootSig;
  DWORD BS_VolID;
  BYTE  BS_VolLab[11];
  BYTE  BS_FilSysType[8];
} tFat32_Bpb;


/* FAT32 FSInfo Sector structure */
typedef struct
{
  DWORD LeadSig;
  BYTE  Reserved1[480];
  DWORD StrucSig;
  DWORD Free_Count;
  DWORD Nxt_Free;
  BYTE  Reserved2[12];
  DWORD TrailSig;
} tFat32_FSInfo;


/* FAT 32-byte Directory Entry structure */
typedef struct
{
  BYTE  Name[11];
  BYTE  Attr;
  BYTE  NTRes;
  BYTE  CrtTimeTenth;
  WORD  CrtTime;
  WORD  CrtDate;
  WORD  LstAccDate;
  WORD  FstClusHI;
  WORD  WrtTime;
  WORD  WrtDate;
  WORD  FstClusLO;
  DWORD FileSize;
} tFatDirEntry;


/* FAT 32-byte Long File Name Directory Entry structure */
typedef struct
{
  BYTE Order;        /* Sequence number for slot        */
  WORD Name0_4[5];   /* First 5 Unicode characters      */
  BYTE Attr;         /* Attributes, always 0x0F         */
  BYTE Reserved;     /* Reserved, always 0x00           */
  BYTE Checksum;     /* Checksum of 8.3 name            */
  WORD Name5_10[6];  /* 6 more Unicode characters       */
  WORD FstClus;      /* First cluster number, must be 0 */
  WORD Name11_12[2]; /* Last 2 Unicode characters       */
} tFatLfnEntry;


/* The file structure stores all the informations about an open file. */
typedef struct
{
  DWORD         Mode;            /* File mode                              */
  DWORD         FirstCluster;    /* The first cluster of the file          */
  tFatDirEntry  DirEntry;        /* The file's directory entry             */

  /* The following fileds refer to the directory that contains the file */
  tFatDirEntry  ParentDirEntry;  /* The parent directory's dir entry       */
  DWORD         DirEntryOffset;  /* Offset of the file's dir entry         */
  int           DirLfnEntries;   /* The number of LFN entries for the file */

  /* The following fields refer to the position into the file */
  SQWORD        TargetPos;       /* The target position for an I/O         */
  SQWORD        FilePos;         /* The current position into the file     */
  DWORD         PrevCluster;     /* The cluster we were before the current */
  DWORD         Cluster;         /* The current cluster                    */
  DWORD         SectorInCluster; /* The sector in the current cluster      */
  DWORD         ByteInSector;    /* The byte in the current sector         */

  /* These are used only if the file is used for a directory search */
  DWORD         FindFlags;
  char          FindFileSpec[FD32_MAX_LFN_LENGTH];
} tFatFile;


#define FAT_BUFFER_USED  0x1
#define FAT_BUFFER_DIRTY 0x2
typedef struct
{
  DWORD  StartingSector;
  DWORD  LastAccess;
  DWORD  Flags;
  BYTE  *Data;
}
tFatBuffer;


/* FIX ME: Max files and max buffers should be read from CONFIG.SYS! */
#define FAT_MAX_FILES 5
#define FAT_MAX_BUFFERS 8
/* This structure stores all informations about a FAT volume.        */
/* It is the P structure (private data) of the FAT driver functions. */
typedef struct FatVolume
{
//  fd32_dev_block_t *Blk;      /* Block device operations                 */
  char DevName[256];
  WORD       SectorsInBuffer; // [KG] count of sectors to buffer
  int        FatType;         /* Can be FAT12, FAT16 or FAT32            */
  DWORD      PartStart;       /* Start of Volume's Partition */
  DWORD      DataClusters;    /* The total number of valid data clusters */
  DWORD      FirstDataSector; /* The first sector of the data region     */
  DWORD      CurDirCluster;   /* The starting cluster of the current dir */
  char       CurDirPath[FD32_MAX_LFN_PATH_LENGTH];
  DWORD      FSI_Free_Count;  /* The count of free clusters              */
  DWORD      FSI_Nxt_Free;    /* The cluster number from which to start  */
                              /* to search for free clusters, if known   */
  DWORD      BufferAccess;    /* The count of accesses into vol buffers  */
  DWORD      BufferMiss;      /* The count of buffer misses              */
  DWORD      BufferHit;       /* The count of buffer hits                */
  tFat32_Bpb Bpb;             /* The BIOS Parameter Block of the volume  */
  tFatFile   Files[FAT_MAX_FILES];
  tFatBuffer Buffer[FAT_MAX_BUFFERS];
} tFatVolume;


/* Finddata block for the internal "find" function */
typedef struct
{
  tFatDirEntry SfnEntry;
  DWORD        EntryOffset;
  int          LfnEntries;
  char         LongName[260];
  char         ShortName[14];
}
tFatFind;

/* Format of the finddata record and search status for DOS style search */
typedef struct
{
  /* Internal fields, BYTE Reserved0[21] */
  BYTE        Drive;
  BYTE        SearchName[11];     /* The first reserved part is 21 bytes long  */
//  tFatVolume far *SearchVol;      /* Volume containing the search directory    */
  BYTE        SearchFlags;    /* Flags specified by findfirst              */
  DWORD       SearchDirStart; /* First cluster of the search directory     */
  DWORD       SearchDirPos;   /* Byte position in the search directory     */
  /* Public fields */
  BYTE  Attr;          /* File attributes                               */
  WORD  MTime;         /* Last modification time in DOS format          */
  WORD  MDate;         /* Last modification date in DOS format          */
  DWORD Size;          /* File size in bytes                            */
  char  Name[13];      /* Null-terminated short file name in ASCII      */
// [KG] due to compatibility to Dos-Searchstruct this information is put into the
//      first 21 reserved Bytes.
//  /* Internal fields, BYTE Reserved1[85] */
//  char        SearchName[FD32_MAX_SFN_PATH_LENGTH]; /* The search template */
} tDosFind;

/* BUFFERS.C - Bufferized I/O support procedures */
int flush_all_buffers(tFatVolume *V);
int read_from_buffer(tFatVolume *V, DWORD Sector, unsigned int *SectorOffset);
//int read_from_buffer (tFatVolume *V, DWORD Sector);

/* DATETIME.C - Files' date and time procedures */
int fat_get_attr(void *P, int Fd, fd32_dev_fs_attr_t *A);
int fat_set_attr(void *P, int Fd, fd32_dev_fs_attr_t *A);

/* DIR.C - Directory management services */
int fat_rmdir(void *P, char *DirName, int UseLfn);
int fat_mkdir(void *P, char *DirName, int UseLfn);
//int fat_chdir(unsigned char *DirName, tFatVolume *Vol, int UseLfn);

/* DIRENTRY.C - Directory entry management procedures */
int find_empty_dir_entries  (tFatVolume *V, int Fd, int NumEntries,
                             DWORD *EntryOffset);
int free_dir_entries        (tFatVolume *V, int Fd, DWORD EntryOffset,
                             int LfnEntries);
int allocate_sfn_dir_entry  (tFatVolume *V, int Fd, tFatDirEntry *D,
                             char *FileName, DWORD *EntryOffset);
int allocate_lfn_dir_entries(tFatVolume *V, int Fd, tFatDirEntry *D,
                             char *Path, char *FileName, DWORD *EntryOffset,
                             int *LfnEntries, WORD Hint);

/* FAT12.C - FAT12 table access procedures */
int fat12_read_entry (tFatVolume *V, DWORD N, int FatNum, DWORD *Value);
int fat12_write_entry(tFatVolume *V, DWORD N, int FatNum, DWORD  Value);

/* FAT16.C - FAT16 table access procedures */
int fat16_read_entry (tFatVolume *V, DWORD N, int FatNum, DWORD *Value);
int fat16_write_entry(tFatVolume *V, DWORD N, int FatNum, DWORD  Value);

/* FAT32.C - FAT32 table access procedures */
int fat32_read_entry (tFatVolume *V, DWORD N, int FatNum, DWORD *Value);
int fat32_write_entry(tFatVolume *V, DWORD N, int FatNum, DWORD  Value);

/* FIND.C - File find procedures */
int find             (tFatVolume *V, int Fd, char *FileSpec,
                      DWORD Flags, tFatFind *FindData, int UseLfn);
int fat_lfn_findfirst(void *P, char *FileSpec, DWORD Flags,
                     fd32_fs_lfnfind_t *FindData);
int fat_lfn_findnext (void *P, int Fd, DWORD Flags,
                     fd32_fs_lfnfind_t *FindData);
int fat_lfn_findclose(void *P, int Fd);
int fat_dos_findfirst(char VolumeNo, char *FileSpec, BYTE Attributes,
                      fd32_fs_dosfind_t *Dta);
//int fat_dos_findfirst(void *P, char *FileSpec, BYTE Attributes,
//                      fd32_fs_dosfind_t *Dta);
int fat_dos_findnext (fd32_fs_dosfind_t *Dta);

/* INIT.C - Mount a FAT volume initializing all its data structures */
//int fat_init_volume(tFatVolume *Vol);
//int free_cluster_count(tFatVolume *V, DWORD *FreeClusters);

/* LFN.C - Long file names support procedures */
BYTE lfn_checksum           (tFatDirEntry *D);
int  split_lfn              (tFatLfnEntry *Slot, tFatDirEntry *D,
                             char *LongName, int *NumSlots);
int  fat_generate_short_name(tFatVolume *Vol, char *Path, char *LongName,
                             char *UniqueShortName, WORD Hint);

/* OPEN.C - File opening and creation procedures */
void split_path(char *FullPath, char *Path, char *Name);
int  fat_open  (void *P, char *FileName, DWORD Mode, WORD Attr,
                WORD AliasHint, int *Result, int UseLfn);
int  reopen_dir(tFatVolume *V, DWORD Mode, DWORD FirstCluster);
int  fat_fflush(void *P, int Fd);
int  fat_close (void *P, int Fd);

/* READ.C - Read a block of data from a file */
int fat_read(void *P, int Fd, BYTE *Buffer, DWORD Size, DWORD *Result);

/* RENAME.C - Rename files and directories */
int fat_rename(void *P, char *OldFullName, char *NewFullName, int UseLfn);

/* SEEK.C - Move file position to the target position specified by lseek */
int fat_lseek(void *P, int Fd, SQWORD Offset, int Origin, SQWORD *Result);
int move_to_targetpos(tFatVolume *V, int Fd, unsigned long BytesToWrite);

/* SHARE.C - Sharing conflict detection */
int check_for_sharing_conflict(tFatVolume *V, int Fd, int FirstOpenFd);
int close_shared_file         (tFatVolume *V, DWORD FirstCluster);

/* SHORTNAM.C - Support code for short file names (8.3 format) */
int  validate_short_file_name   (char *Dest, char *Source);
void absolute_name_to_short_name(char *Dest, char *Source);
void short_name_to_absolute_name(char *source, char *dest);

/* SUPPORT.C - Small support procedures */
tFatFile *validate_fd            (tFatVolume *V, DWORD Fd);
void      update_date_and_time   (WORD *Time, WORD *Date, BYTE *Hund);
DWORD     clusters_amount        (DWORD N, tFat32_Bpb *Bpb);
DWORD     first_root_dir_sector  (tFat32_Bpb *Bpb);
DWORD     first_sector_of_cluster(DWORD N, tFatVolume *Vol);

int fat_free_disk_space(tFatVolume *V, DWORD *SectorsPerCluster,
                        DWORD *FreeClusters, DWORD *BytesPerSector,
                        DWORD *TotalClusters);


/* UNLINK.C - File and clusters unlinking procedures */
int unlink_clusters(tFatVolume *V, DWORD FromCluster);
int fat_unlink     (void *P, char *FileName, DWORD Flags, int UseLfn);

/* WRITE.C - Write a block of data into a file truncating or extending it */
int fat_write(void *P, int Fd, BYTE *Buffer, DWORD Size, DWORD *Result);

#endif

