/**************************************************************************
 * FreeDOS 32 FAT Driver                                                  *
 * by Salvo Isaja                                                         *
 *                                                                        *
 * Copyright (C) 2001-2002, Salvatore Isaja                               *
 *                                                                        *
 * This is "fat12.c" - Low level internal services to read/write entries  *
 *                     of the file allocation table in FAT12 format       *
 *                                                                        *
 *                                                                        *
 * This file is part of the FreeDOS 32 FAT Driver.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is free software; you can redistribute it    *
 * and/or modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2 of the  *
 * License, or (at your option) any later version.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is distributed in the hope that it will be   *
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 * GNU General Public License for more details.                           *
 *                                                                        *
 * You should have received a copy of the GNU General Public License      *
 * along with the FreeDOS 32 FAT Driver; see the file COPYING;            *
 * if not, write to the Free Software Foundation, Inc.,                   *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                *
 **************************************************************************/

#include "fat1.h"


#if 0
static unsigned short get_word_from_cache(tFatFile *F, unsigned long Sector, unsigned long EntryOffset, tFatVolume *Vol)
{
  int Offset = (Sector - F->FatCache.FirstSector) * 512 + EntryOffset;
  //printf("Offset : %li\n",Offset);

  if ((Offset < 0) || (Offset >= 4090) || (F->FatCache.FirstSector == 0xFFFFFFFF))
  /* cache miss */
  {
    #ifdef DEBUG
    printf("FAT12 Cache Miss\n");
    #endif
    F->FatCache.FirstSector = Sector;
    biosdisk_read_sectors_lba(&Vol->DiskInfo, Vol->StartingSector + Sector, 8192/512, F->FatCache.Data);
    Offset = (Sector - F->FatCache.FirstSector) * 512 + EntryOffset;
  }

  return *((unsigned short *) &F->FatCache.Data[Offset]);
}
#endif


/* The location of a valid cluster N into the FAT12 FAT       */
/* Returns 0 on success, or a negative error code on failure. */
/* Called by fat12_read_entry and fat12_write_entry.          */
static int fat12_cluster_entry(tFatVolume *V, DWORD N, int FatNum,
                               DWORD *Sector, DWORD *EntryOffset)
{
  DWORD FATSz, FATOffset;

  if (N > V->DataClusters + 1) return FAT_RET_INVALID_CLUSTER;
  /* Multiply by 1.5 rounding down */
  FATOffset = N + (N / 2);
  if (V->Bpb.BPB_FATSz16 != 0) FATSz = V->Bpb.BPB_FATSz16;
                          else FATSz = V->Bpb.BPB_FATSz32;

  *Sector      = FatNum * FATSz + V->Bpb.BPB_ResvdSecCnt
               + FATOffset / V->Bpb.BPB_BytsPerSec;
  *EntryOffset = FATOffset % V->Bpb.BPB_BytsPerSec;
  return 0;
}


/* Reads the value of the specified cluster entry in a FAT12 FAT.      */
/* Returns 0 on success, or a negative error code on failure.          */
/* Called by free_cluster_count (init.c), first_free_cluster (seek.c), */
/* advance_byte_position (seek.c), truncate_or_extend (write.c) and    */
/* unlink_clusters (unlink.c).                                         */
int fat12_read_entry(tFatVolume *V, DWORD N, int FatNum, DWORD *Value)
{
  /* Since the size of a FAT12 entry is not multiple of 2, */
  /* an entry could span a sector boundary.                */
  int   NumBuf1, NumBuf2;
  int   Res;
  DWORD Sector, EntryOffset;
unsigned int SectorOffset,SectorOffset1,SectorOffset2;

  Res = fat12_cluster_entry(V, N, FatNum, &Sector, &EntryOffset);
  if (Res) return Res;

  /* Check if the entry doesn't span a sector boundary */
  if (EntryOffset < V->Bpb.BPB_BytsPerSec - 1)
  {
    if ((NumBuf1 = read_from_buffer(V, Sector, &SectorOffset)) < 0) return NumBuf1;
    *Value = *((WORD *) &V->Buffer[NumBuf1].Data[EntryOffset+SectorOffset]);
  }
  else
  {
    if ((NumBuf1 = read_from_buffer(V, Sector, &SectorOffset1))     < 0) return NumBuf1;
    if ((NumBuf2 = read_from_buffer(V, Sector + 1, &SectorOffset2)) < 0) return NumBuf2;
    *Value = V->Buffer[NumBuf1].Data[EntryOffset+SectorOffset1]
           + (((unsigned int)V->Buffer[NumBuf2].Data[0+SectorOffset2]) << 8);
  }
  /* If the entry number is odd, we need the highest 12 bits of the 16-bit */
  /* Value. If the entry number is even, we need the lowest 12 bits.       */
  if (N & 0x0001) *Value >>= 4;
             else *Value &= 0x0FFF;
  return 0;
}


/* Writes the value of the specified cluster entry in a FAT12 FAT.          */
/* Returns 0 on success, or a negative error code on failure.               */
/* Called by allocate_and_link_new_cluster (seek.c), allocate_first_cluster */
/* (seek.c), truncate_or_extend (write.c) and unlink_clusters (unlink.c).   */
int fat12_write_entry(tFatVolume *V, DWORD N, int FatNum, DWORD Value)
{
  /* Since the size of a FAT12 entry is not multiple of 2, */
  /* an entry could span a sector boundary.                */
  int   NumBuf1, NumBuf2;
  int   Res;
  DWORD Sector, EntryOffset;
unsigned int SectorOffset1,SectorOffset2;

  Res = fat12_cluster_entry(V, N, FatNum, &Sector, &EntryOffset);
  if (Res) return Res;

  /* Check if the entry doesn't span a sector boundary */
  if (EntryOffset < V->Bpb.BPB_BytsPerSec - 1)
  {
    if ((NumBuf1 = read_from_buffer(V, Sector, &SectorOffset1)) < 0) return NumBuf1;
    /* If the entry number is odd, we need to place the entry value in    */
    /* the highest 12 bits of a 16-bit SecBuff entry. If the entry number */
    /* is even, we need to place the entry value in the lower 12 bits.    */
    if (N & 0x0001)
    {
      Value <<= 4;
      *((WORD *) &V->Buffer[NumBuf1].Data[EntryOffset+SectorOffset1]) &= 0x000F;
    }
    else
    {
      Value &= 0x0FFF;
      *((WORD *) &V->Buffer[NumBuf1].Data[EntryOffset+SectorOffset1]) &= 0xF000;
    }
    *((WORD *) &V->Buffer[NumBuf1].Data[EntryOffset+SectorOffset1]) |= Value;
    V->Buffer[NumBuf1].Flags |= FAT_BUFFER_DIRTY;
  }
  else
  {
    if ((NumBuf1 = read_from_buffer(V, Sector, &SectorOffset1))     < 0) return NumBuf1;
    if ((NumBuf2 = read_from_buffer(V, Sector + 1, &SectorOffset2)) < 0) return NumBuf2;
    /* If the entry number is odd, we need to place the entry value in    */
    /* the highest 12 bits of a 16-bit SecBuff entry. If the entry number */
    /* is even, we need to place the entry value in the lower 12 bits.    */
    if (N & 0x0001)
    {
      Value <<= 4;
      V->Buffer[NumBuf1].Data[EntryOffset+SectorOffset1] &= 0x0F; /* LSB */
      V->Buffer[NumBuf2].Data[0+SectorOffset2]            = 0x00; /* MSB */
    }
    else
    {
      Value &= 0x0FFF;
      V->Buffer[NumBuf1].Data[EntryOffset+SectorOffset1] = Value & 0xFF; /* LSB */
      V->Buffer[NumBuf2].Data[0+SectorOffset2]          &= 0xF0; /* MSB */
    }
    V->Buffer[NumBuf1].Data[EntryOffset+SectorOffset1] |= Value & 0xFF; /* LSB */
    V->Buffer[NumBuf2].Data[0+SectorOffset2]           |= Value >> 8;   /* MSB */
    V->Buffer[NumBuf1].Flags |= FAT_BUFFER_DIRTY;
    V->Buffer[NumBuf2].Flags |= FAT_BUFFER_DIRTY;
  }
  return 0;
}
