/* FreeDOS32 FAT Driver
 * Test program for file opening, reading and closing (fat_open/read/close)
 * by Salvo Isaja
 */

#include <stdio.h>
#include "fat1.h"

/* BUFFERS.C - Bufferized I/O support procedures */
#include "buffers.c"
/* DATETIME.C - Files' date and time procedures */
#include "datetime.c"
/* DIR.C - Directory management services */
#include "dir.c"
/* DIRENTRY.C - Directory entry management procedures */
#include "direntry.c"
/* FAT12.C - FAT12 table access procedures */
#include "fat12.c"
/* FAT16.C - FAT16 table access procedures */
#include "fat16.c"
/* FAT32.C - FAT32 table access procedures */
#include "fat32.c"
/* FIND.C - File find procedures */
#include "find.c"
/* INIT.C - Mount a FAT volume initializing all its data structures */
#include "init.c"
/* LFN.C - Long file names support procedures */
#include "lfn.c"
/* OPEN.C - File opening and creation procedures */
#include "open.c"
/* READ.C - Read a block of data from a file */
#include "read.c"
/* RENAME.C - Rename files and directories */
#include "rename.c"
/* SEEK.C - Move file position to the target position specified by lseek */
#include "seek.c"
/* SHARE.C - Sharing conflict detection */
#include "share.c"
/* SHORTNAM.C - Support code for short file names (8.3 format) */
#include "shortnam.c"
/* SUPPORT.C - Small support procedures */
#include "support.c"
/* UNLINK.C - File and clusters unlinking procedures */
#include "unlink.c"
/* WRITE.C - Write a block of data into a file truncating or extending it */
#include "write.c"
#include "unicode.c"
#include "wildcard.c"

#define BUFSIZE 32768

int main()
{
  tFatVolume    V;
  int           Fd;
  int           Res;
  unsigned long NumRead;
  unsigned char Buffer[BUFSIZE];

  fat_init_volume(&V);

  /* Pass the following parameters to fat_open():
   * - ASCIZ file name (path allowed)
   * - Pointer to an int variable to store the handle allocated for the file
   * - Pointer to FAT Volume description structure
   * - Opening mode flags
   * - Attributes to assign to the file if it's created
   * - Numeric hint for short name alias generation
   * - Pointer to an int variable to store the result for the action taken
   * - Set to nonzero if you want to use Long File Names
   */
  Fd = fat_open(&V, "\\djgpp\\djsetup.html", FD32_OPEN_ACCESS_READ | FD32_OPEN_EXIST_OPEN,
                 FD32_ATTR_NONE, 0, NULL, 1);
  if (Fd < 0)
  {
    printf("Failed due to error %i\n",Fd);
    return Fd;
  }
  /* Pass the following parameters to fat_read():
   * - Handle of the file to read from
   * - Pointer to FAT Volume description structure
   * - Pointer to a buffer of bytes (unsigned chars) to store the bytes read
   * - Size of the block of data to read
   * - Pointer to a variable to store the number of bytes actually read
   *   (may be NULL if you are not interested in it)
   */
  Res = fat_read(&V, Fd, Buffer, BUFSIZE, &NumRead);
  if (Res)
  {
    printf("Failed due to error %i\n",Res);
    return Res;
  }
  printf("%lu bytes read.\n",NumRead);
  /* Pass the following parameters to fat_close():
   * - Handle of the file to be closed
   * - Pointer to FAT Volume description structure
   */
  Res = fat_close(&V, Fd);
  if (Res)
  {
    printf("Failed due to error %i\n",Res);
    return Res;
  }
  return 0;
}

